
#include "GraspMemory.h"

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <SimoxUtility/algorithm/string.h>

#include <RobotAPI/libraries/armem/core/error.h>
#include <RobotAPI/libraries/armem/server/MemoryRemoteGui.h>

#include <RobotAPI/libraries/GraspingUtility/aron/GraspCandidate.aron.generated.h>
#include <RobotAPI/libraries/GraspingUtility/GraspCandidateVisu.h>
#include <RobotAPI/libraries/GraspingUtility/aron_conversions.h>
#include <RobotAPI/libraries/armem_grasping/aron/KnownGraspCandidate.aron.generated.h>


namespace armarx::armem::server::grasp
{
    armarx::PropertyDefinitionsPtr GraspMemory::createPropertyDefinitions()
    {
        armarx::PropertyDefinitionsPtr defs = new ComponentPropertyDefinitions(getConfigIdentifier());

        defs->topic(debugObserver);

        defs->optional(memoryName, "memory.Name", "Name of this memory server.");

        return defs;
    }


    std::string GraspMemory::getDefaultName() const
    {
        return "GraspMemory";
    }

    GraspMemory::GraspMemory() :
        knownGraspProviderSegment(iceAdapter())
    {
    }

    void GraspMemory::onInitComponent()
    {
        workingMemory().name() = memoryName;

        workingMemory().addCoreSegment("GraspCandidate",
                                     armarx::grasping::arondto::GraspCandidate::ToAronType());
        workingMemory().addCoreSegment("BimanualGraspCandidate",
                                     armarx::grasping::arondto::BimanualGraspCandidate::ToAronType());
        workingMemory().addCoreSegment("KnownGraspCandidate",
                                     armarx::armem::grasping::arondto::KnownGraspInfo::ToAronType());

        knownGraspProviderSegment.init();
    }

    void GraspMemory::onConnectComponent()
    {

//        createRemoteGuiTab();
//        RemoteGui_startRunningTask();

    }

    void GraspMemory::onDisconnectComponent()
    {
    }

    void GraspMemory::onExitComponent()
    {
    }



    // WRITING

    armem::data::CommitResult GraspMemory::commit(const armem::data::Commit& commit, const Ice::Current&)
    {
        // This function is overloaded to trigger the remote gui rebuild.
        armem::data::CommitResult result = ReadWritePluginUser::commit(commit);
        gui.tab.rebuild = true;
        return result;
    }


    // READING

    // Inherited from Plugin



    // REMOTE GUI

    void GraspMemory::createRemoteGuiTab()
    {
        //add all instances to map
        workingMemory().getCoreSegment("GraspCandidate").forEachInstance([this](const auto & instance)
        {
            if (gui.candidatesToDraw.find(instance.id().str()) == gui.candidatesToDraw.end()){
                gui.candidatesToDraw[instance.id().str()] = false;
            }
        });

        using namespace armarx::RemoteGui::Client;
        GridLayout root;
        {
            gui.tab.selectAll.setLabel("Select All");
            gui.tab.deselectAll.setLabel("Deselect All");
            int row = 0;

            // bimanual candidates are not available for visualization for now
//            {
//                root.add(Label("CoreSegment"), Pos{row, 0});
//                row++;
//                std::vector<std::string> options = workingMemory().getCoreSegmentNames();
//                if (options.empty())
//                {
//                    options.push_back("<None>");
//                }
//                gui.tab.selectCoreSegment.setOptions(options);
//                if (gui.coreSegIndex >= 0 && gui.coreSegIndex < static_cast<int>(options.size()))
//                {
//                    gui.tab.selectCoreSegment.setIndex(gui.coreSegIndex);
//                    gui.coreSegment = options[gui.coreSegIndex];
//                }
//                root.add(gui.tab.selectCoreSegment, Pos{row,0});
//                row++;
//            }
            if(workingMemory().hasCoreSegment(gui.coreSegment))
            {
                {
                    root.add(Label("Provider"), Pos{row, 0});
                    row++;
                    std::vector<std::string> options = workingMemory().getCoreSegment(gui.coreSegment).getProviderSegmentNames();
                    if (options.empty())
                    {
                        options.push_back("<None>");
                    }
                    else
                    {
                        options.insert(options.begin(), "<All>");
                    }

                    gui.tab.selectProvider.setOptions(options);

                    if (gui.providerIndex >= 0 && gui.providerIndex < static_cast<int>(options.size()))
                    {
                        gui.tab.selectProvider.setIndex(gui.providerIndex);
                        gui.provider = options[gui.providerIndex];
                    }
                    root.add(gui.tab.selectProvider, Pos{row, 0});
                    row++;
                }

                if(gui.provider == "<All>"
                        || workingMemory().getCoreSegment(gui.coreSegment).hasProviderSegment(gui.provider))
                {
                    std::vector<MemoryID> providers;
                    {
                        root.add(Label("Entity"), Pos{row,0});
                        row++;
                        std::vector<std::string> options;

                        if (gui.provider == "<All>")
                        {
                            workingMemory().getCoreSegment(gui.coreSegment).forEachProviderSegment([&providers](const auto & provider)
                            {
                                providers.push_back(provider.id());
                            });
                        }
                        else
                        {
                            providers.push_back(MemoryID(memoryName, gui.coreSegment, gui.provider));

                        }
                        for (MemoryID & provider : providers)
                        {
                            for(std::string & entity : workingMemory().getProviderSegment(provider).getEntityNames())
                            {
                                options.push_back(entity);
                            }
                        }
                        if (options.empty())
                        {
                            options.push_back("<None>");
                        }
                        else
                        {
                            options.insert(options.begin(), "<All>");
                        }
                        gui.tab.selectEntity.setOptions(options);
                        if (gui.entityIndex >= 0 && gui.entityIndex < static_cast<int>(options.size()))
                        {
                            gui.tab.selectEntity.setIndex(gui.entityIndex);
                            gui.entity = options[gui.entityIndex];
                        }
                        root.add(gui.tab.selectEntity, Pos{row, 0});
                        row++;
                    }

                    if( gui.entity == "<All>"
                            || (gui.provider == "<All>" && workingMemory().getCoreSegment(gui.coreSegment).hasEntity(gui.entity))
                            || workingMemory().getCoreSegment(gui.coreSegment).getProviderSegment(gui.provider).hasEntity(gui.entity))
                    {
                        std::vector<MemoryID> entities;
                        {
                            root.add(Label("Snapshot"), Pos{row, 0});
                            row++;
                            std::vector<std::string> options;
                            for (MemoryID & provider : providers)
                            {
                                if (gui.entity == "<All>")
                                {
                                    workingMemory().getProviderSegment(provider).forEachEntity([&entities](auto & entity)
                                    {
                                        entities.push_back(entity.id());
                                    });
                                }
                                else
                                {
                                    if (workingMemory().getProviderSegment(provider).hasEntity(gui.entity))
                                    {
                                        entities.push_back(MemoryID(memoryName, gui.coreSegment, provider.providerSegmentName, gui.entity));
                                    }
                                }
                            }

                            for (MemoryID & entity : entities)
                            {
                                workingMemory().getEntity(entity).forEachSnapshot([this, &options](const auto & snapshot)
                                {
                                    std::string time = armem::toDateTimeMilliSeconds(snapshot.time());
                                    gui.timeMap[time] = snapshot.time();
                                    options.push_back(time);
                                });
                            }

                            if (options.empty())
                            {
                                options.push_back("<None>");
                            }
                            else
                            {
                                options.insert(options.begin(), "<All>");
                            }
                            gui.tab.selectSnapshot.setOptions(options);
                            if (gui.snapshotIndex >= 0 && gui.snapshotIndex < static_cast<int>(options.size()))
                            {
                                gui.tab.selectSnapshot.setIndex(gui.snapshotIndex);
                                if(options[gui.snapshotIndex] != "<None>")
                                {
                                    gui.snapshot = options[gui.snapshotIndex];
                                }
                            }
                            root.add(gui.tab.selectSnapshot, Pos{row, 0});
                            row++;


                        }
                        bool comboExists;
                        if (gui.snapshot == "<All>")
                        {
                            comboExists = true;
                        }
                        else
                        {
                            for (MemoryID & entity : entities)
                            {
                                if (workingMemory().getEntity(entity).hasSnapshot(gui.timeMap.at(gui.snapshot)))
                                {
                                    comboExists = true;
                                }
                            }
                        }
                        if(comboExists)
                        {
                            root.add(Label("Instances"), Pos{row,0});
                            row++;
                            root.add(gui.tab.selectAll, Pos{row, 0});
                            row++;
                            root.add(gui.tab.deselectAll, Pos{row, 0});
                            row++;
                            gui.tab.checkInstances.clear();
                            std::vector<MemoryID> snapshots;
                            for (MemoryID & entity : entities)
                            {
                                if(gui.snapshot == "<All>")
                                {
                                    workingMemory().getEntity(entity).forEachSnapshot([&snapshots](auto & snapshot)
                                    {
                                        snapshots.push_back(snapshot.id());
                                    });
                                }
                                else
                                {
                                    if (workingMemory().getEntity(entity).hasSnapshot(gui.timeMap.at(gui.snapshot)))
                                    {
                                        snapshots.push_back(workingMemory().getEntity(entity).getSnapshot(gui.timeMap.at(gui.snapshot)).id());
                                    }
                                }
                            }

                            for (MemoryID & snapshot : snapshots)
                            {
                                workingMemory().getSnapshot(snapshot).forEachInstance([this, &row, &root](const auto & instance)
                                {
                                   root.add(Label(instance.id().str()) , Pos{row, 0});
                                   gui.tab.checkInstances[instance.id().str()] = CheckBox();
                                   if(gui.candidatesToDraw.at(instance.id().str()))
                                   {
                                       gui.tab.checkInstances.at(instance.id().str()).setValue(true);
                                   }
                                   root.add(gui.tab.checkInstances[instance.id().str()], Pos{row, 1});
                                   row++;
                                });
                            }
                        }
                        else
                        {
                            root.add(Label("Instances"), Pos{row,0});
                            row++;
                            root.add(gui.tab.selectAll, Pos{row, 0});
                            row++;
                            root.add(gui.tab.deselectAll, Pos{row, 0});
                            row++;
                            gui.tab.checkInstances.clear();
                        }
                    }
                    else
                    {
                        gui.tab.checkInstances.clear();

                        std::vector<std::string> options = {"<None>"};

                        gui.tab.selectSnapshot.setOptions(options);

                        root.add(Label("Snapshot"), Pos{row, 0});
                        row++;
                        root.add(gui.tab.selectSnapshot, Pos{row, 0});
                        row++;
                        root.add(gui.tab.selectAll, Pos{row, 0});
                        row++;
                        root.add(gui.tab.deselectAll, Pos{row, 0});
                        row++;
                    }
                }
                else
                {
                    gui.tab.checkInstances.clear();

                    std::vector<std::string> options = {"<None>"};

                    gui.tab.selectEntity.setOptions(options);

                    root.add(Label("Entity"), Pos{row,0});
                    row++;
                    root.add(gui.tab.selectEntity, Pos{row, 0});
                    row++;

                    gui.tab.selectSnapshot.setOptions(options);

                    root.add(Label("Snapshot"), Pos{row, 0});
                    row++;
                    root.add(gui.tab.selectSnapshot, Pos{row, 0});
                    row++;
                    root.add(gui.tab.selectAll, Pos{row, 0});
                    row++;
                    root.add(gui.tab.deselectAll, Pos{row, 0});
                    row++;
                }
            }
            else {
                gui.tab.checkInstances.clear();

                std::vector<std::string> options = {"<None>"};
                gui.tab.selectProvider.setOptions(options);

                root.add(Label("Provider"), Pos{row, 0});
                row++;
                root.add(gui.tab.selectProvider, Pos{row, 0});
                row++;

                gui.tab.selectEntity.setOptions(options);

                root.add(Label("Entity"), Pos{row,0});
                row++;
                root.add(gui.tab.selectEntity, Pos{row, 0});
                row++;

                gui.tab.selectSnapshot.setOptions(options);

                root.add(Label("Snapshot"), Pos{row, 0});
                row++;
                root.add(gui.tab.selectSnapshot, Pos{row, 0});
                row++;
                root.add(gui.tab.selectAll, Pos{row, 0});
                row++;
                root.add(gui.tab.deselectAll, Pos{row, 0});
                row++;

            }
        }


        RemoteGui_createTab(getName(), root, &gui.tab);
        gui.tab.rebuild = false;
    }


    void GraspMemory::RemoteGui_update()
    {
        armarx::grasping::GraspCandidateVisu visu;

//        if (gui.tab.selectCoreSegment.hasValueChanged())
//        {
//            gui.coreSegment = gui.tab.selectCoreSegment.getValue();
//            gui.coreSegIndex = gui.tab.selectCoreSegment.getIndex();
//            gui.providerIndex = 0;
//            gui.entityIndex = 0;
//            gui.snapshotIndex = 0;
//            gui.tab.rebuild = true;
//        }

        if (gui.tab.selectProvider.hasValueChanged())
        {
            gui.provider = gui.tab.selectProvider.getValue();
            gui.providerIndex = gui.tab.selectProvider.getIndex();
            gui.entityIndex = 0;
            gui.snapshotIndex = 0;
            gui.tab.rebuild = true;
        }

        if (gui.tab.selectEntity.hasValueChanged())
        {
            gui.entity = gui.tab.selectEntity.getValue();
            gui.entityIndex = gui.tab.selectEntity.getIndex();
            gui.snapshotIndex = 0;
            gui.tab.rebuild = true;
        }

        if (gui.tab.selectSnapshot.hasValueChanged())
        {
            gui.snapshot = gui.tab.selectSnapshot.getValue();
            gui.snapshotIndex = gui.tab.selectSnapshot.getIndex();
            gui.tab.rebuild = true;
        }


        if(gui.tab.selectAll.wasClicked())
        {
            for(auto & element : gui.tab.checkInstances)
            {
                element.second.setValue(true);
            }
        }

        if(gui.tab.deselectAll.wasClicked())
        {
            for(auto & element : gui.tab.checkInstances)
            {
                element.second.setValue(false);
            }
        }

        for(auto & element : gui.tab.checkInstances)
        {
            gui.candidatesToDraw.at(element.first) = element.second.getValue();
        }

        std::map<std::string, viz::Layer> reachableLayers;
        std::map<std::string, viz::Layer> unreachableLayers;

        for(auto & element : gui.candidatesToDraw )
        {
            std::string entityName = armem::MemoryID::fromString(element.first).entityName;

            if (reachableLayers.find(entityName) == reachableLayers.end())
            {
                reachableLayers[entityName] = arviz.layer(entityName + "_reachable");
                unreachableLayers[entityName] = arviz.layer(entityName + "_unreachable");
            }

            //draw candidates
            if (element.second)
            {
                armarx::grasping::GraspCandidate candidate;

                armarx::grasping::arondto::GraspCandidate aronTransform;

                armem::wm::EntityInstance instance = workingMemory().getInstance(armem::MemoryID::fromString(element.first));

                aronTransform.fromAron(workingMemory().getInstance(armem::MemoryID::fromString(element.first)).data());

                fromAron(aronTransform, candidate);

                visu.visualize(element.first, candidate, reachableLayers.at(entityName), unreachableLayers.at(entityName));
            }
        }

        std::vector<viz::Layer> layers;
        for(auto & [entityName, layer] : reachableLayers)
        {
            layers.push_back(layer);
            layers.push_back(unreachableLayers.at(entityName));
        }
        arviz.commit(layers);

        if (gui.tab.rebuild)
        {
            createRemoteGuiTab();
        }
    }

}
